import logging

from motile_toolbox.utils import relabel_segmentation
from motile_toolbox.visualization import to_napari_tracks_layer
from napari import Viewer
from napari.layers import Labels, Tracks
from qtpy.QtCore import Signal
from qtpy.QtWidgets import (
    QLabel,
    QVBoxLayout,
    QWidget,
)
from superqt.utils import thread_worker

from motile_plugin.backend.motile_run import MotileRun
from motile_plugin.backend.solve import solve

from .run_editor import RunEditor
from .run_viewer import RunViewer
from .runs_list import RunsList

logger = logging.getLogger(__name__)


class MotileWidget(QWidget):
    """The main widget for the motile napari plugin. Coordinates sub-widgets
    and calls the back-end motile solver.
    """

    # A signal for passing events from the motile solver to the run view widget
    # To provide updates on progress of the solver
    solver_update = Signal()

    def __init__(self, viewer: Viewer):
        super().__init__()
        self.viewer: Viewer = viewer

        # Declare napari layers for displaying outputs (managed by the widget)
        self.output_seg_layer: Labels | None = None
        self.tracks_layer: Tracks | None = None

        # Create sub-widgets and connect signals
        self.edit_run_widget = RunEditor(self.viewer)
        self.edit_run_widget.start_run.connect(self._generate_tracks)

        self.view_run_widget = RunViewer()
        self.view_run_widget.edit_run.connect(self.edit_run)
        self.view_run_widget.hide()
        self.solver_update.connect(self.view_run_widget.solver_event_update)

        self.run_list_widget = RunsList()
        self.run_list_widget.view_run.connect(self.view_run_napari)

        # Create main layout
        main_layout = QVBoxLayout()
        main_layout.addWidget(self._title_widget())
        main_layout.addWidget(self.view_run_widget)
        main_layout.addWidget(self.edit_run_widget)
        main_layout.addWidget(self.run_list_widget)
        self.setLayout(main_layout)

    def remove_napari_layers(self) -> None:
        """Remove the currently stored layers from the napari viewer, if present"""
        if (
            self.output_seg_layer
            and self.output_seg_layer in self.viewer.layers
        ):
            self.viewer.layers.remove(self.output_seg_layer)
        if self.tracks_layer and self.tracks_layer in self.viewer.layers:
            self.viewer.layers.remove(self.tracks_layer)

    def update_napari_layers(self, run: MotileRun) -> None:
        """Remove the old napari layers and update them according to the run output.
        Will create new segmentation and tracks layers and add them to the viewer.

        Args:
            run (MotileRun): The run outputs to visualize in napari.
        """
        # Remove old layers if necessary
        self.remove_napari_layers()

        # Create new layers
        if run.output_segmentation is not None:
            self.output_seg_layer = Labels(
                run.output_segmentation[:, 0], name=run.run_name + "_seg"
            )
        else:
            self.output_seg_layer = None

        if run.tracks is None or run.tracks.number_of_nodes() == 0:
            self.tracks_layer = None
        else:
            track_data, track_props, track_edges = to_napari_tracks_layer(
                run.tracks
            )
            self.tracks_layer = Tracks(
                track_data,
                properties=track_props,
                graph=track_edges,
                name=run.run_name + "_tracks",
                tail_length=3,
            )

        # Add layers to the viewer
        if self.tracks_layer is not None:
            self.viewer.add_layer(self.output_seg_layer)
            self.viewer.add_layer(self.tracks_layer)

    def view_run_napari(self, run: MotileRun) -> None:
        """Populates the run viewer and the napari layers with the output
        of the provided run.

        Args:
            run (MotileRun): The run to view
        """
        self.view_run_widget.update_run(run)
        self.edit_run_widget.hide()
        self.view_run_widget.show()
        self.update_napari_layers(run)

    def edit_run(self, run: MotileRun | None):
        """Create or edit a new run in the run editor. Also removes solution layers
        from the napari viewer.
        Args:
            run (MotileRun | None): Initialize the new run with the parameters and name
                from this run. If not provided, uses the SolverParams default values.
        """
        self.view_run_widget.hide()
        self.edit_run_widget.show()
        if run:
            self.edit_run_widget.new_run(run)
        self.run_list_widget.runs_list.clearSelection()
        self.remove_napari_layers()

    def _generate_tracks(self, run: MotileRun) -> None:
        """Called when we start solving a new run. Switches from run editor to run viewer
        and starts solving of the new run in a separate thread to avoid blocking

        Args:
            run (MotileRun): Start solving this motile run.
        """
        run.status = "initializing"
        self.run_list_widget.add_run(run, select=True)
        worker = self.solve_with_motile(run)
        worker.returned.connect(self._on_solve_complete)
        worker.start()

    @thread_worker
    def solve_with_motile(self, run: MotileRun) -> MotileRun:
        """Runs the solver and relabels the segmentation to match
        the solution graph.
        Emits: self.solver_event when the solver provides an update
        (will be emitted from the thread, which is why it needs to be an
        event and not just a normal function callback)

        Args:
            run (MotileRun): A run with name, parameters, and input segmentation,
                but not including the output graph or segmentation.

        Returns:
            MotileRun: The provided run with the output graph and segmentation included.
        """
        run.tracks = solve(
            run.solver_params,
            run.input_segmentation,
            lambda event_data: self._on_solver_event(run, event_data),
        )
        run.output_segmentation = relabel_segmentation(
            run.tracks, run.input_segmentation
        )
        return run

    def _on_solver_event(self, run: MotileRun, event_data: dict) -> None:
        """Parse the solver event and set the run status and gap accordingly.
        Also emits a solver_update event to tell the run viewer to update.
        Note: This will simply tell the run viewer to refresh its plot and
        status. If the run viewer is not viewing this run, it will refresh
        anyways, which is pointless but not harmful.

        Args:
            run (MotileRun): The run that the solver is working on
            event_data (dict): The solver event data from ilpy.EventData
        """
        event_type = event_data["event_type"]
        if (
            event_type in ["PRESOLVE", "PRESOLVEROUND"]
            and run.status != "presolving"
        ):
            run.status = "presolving"
            run.gaps = (
                []
            )  # try this to remove the weird initial gap for gurobi
            self.solver_update.emit()
        elif event_type in ["MIPSOL", "BESTSOLFOUND"]:
            run.status = "solving"
            gap = event_data["gap"]
            run.gaps.append(gap)
            self.solver_update.emit()

    def _on_solve_complete(self, run: MotileRun) -> None:
        """Called when the solver thread returns. Updates the run status to done
        and tells the run viewer to update.

        Args:
            run (MotileRun): The completed run
        """
        run.status = "done"
        self.solver_update.emit()
        self.view_run_napari(run)

    def _title_widget(self) -> QWidget:
        """Create the title and intro paragraph widget, with links to docs

        Returns:
            QWidget: A widget introducing the motile plugin and linking to docs
        """
        richtext = r"""<h3>Tracking with Motile</h3>
        <p>This plugin uses the
        <a href="https://funkelab.github.io/motile/">motile</a> library to
        track objects with global optimization. See the
        <a href="https://funkelab.github.io/motile-napari-plugin/">user guide</a>
        for a tutorial to the plugin functionality."""
        label = QLabel(richtext)
        label.setWordWrap(True)
        return label



from __future__ import annotations

import logging
from typing import TYPE_CHECKING
from warnings import warn

import magicgui.widgets
import napari.layers
import numpy as np
from qtpy.QtCore import Signal
from qtpy.QtWidgets import (
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from motile_plugin.backend.motile_run import MotileRun

from .params_editor import SolverParamsEditor

if TYPE_CHECKING:
    import napari

logger = logging.getLogger(__name__)


class RunEditor(QGroupBox):
    start_run = Signal(MotileRun)

    def __init__(self, viewer: napari.Viewer):
        """A widget for editing run parameters and starting solving.
        Has to know about the viewer to get the input segmentation
        from the current layers.

        Args:
            viewer (napari.Viewer): The napari viewer that the editor should
                get the input segmentation from.
        """
        super().__init__(title="Run Editor")
        self.viewer = viewer
        self.solver_params_widget = SolverParamsEditor()
        self.run_name: QLineEdit
        self.layer_selection_box: magicgui.widgets.Widget

        # Generate Tracks button
        generate_tracks_btn = QPushButton("Run Tracking")
        generate_tracks_btn.clicked.connect(self.emit_run)
        generate_tracks_btn.setToolTip(
            "Might take minutes or longer for larger samples."
        )

        main_layout = QVBoxLayout()
        main_layout.addWidget(self._run_widget())
        main_layout.addWidget(self._labels_layer_widget())
        main_layout.addWidget(self.solver_params_widget)
        main_layout.addWidget(generate_tracks_btn)
        self.setLayout(main_layout)

    def _labels_layer_widget(self) -> QWidget:
        """Create the widget to select the input layer. Uses magicgui,
        but explicitly connects to the viewer layers events to keep it synced.

        Returns:
            QWidget: A dropdown select with all the labels layers in layers
                and a refresh button to sync with napari.
        """
        layer_group = QWidget()
        layer_layout = QHBoxLayout()
        layer_layout.setContentsMargins(0, 0, 0, 0)
        label = QLabel("Input Layer:")
        layer_layout.addWidget(label)
        label.setToolTip(
            "Select the labels layer you want to use for tracking"
        )

        # # Layer selection combo box
        self.layer_selection_box = magicgui.widgets.create_widget(
            annotation=napari.layers.Labels
        )
        layers_events = self.viewer.layers.events
        layers_events.inserted.connect(self.layer_selection_box.reset_choices)
        layers_events.removed.connect(self.layer_selection_box.reset_choices)
        layers_events.reordered.connect(self.layer_selection_box.reset_choices)

        qlayer_select = self.layer_selection_box.native

        size_policy = qlayer_select.sizePolicy()
        size_policy.setHorizontalPolicy(QSizePolicy.MinimumExpanding)
        qlayer_select.setSizePolicy(size_policy)
        layer_layout.addWidget(qlayer_select)

        layer_group.setLayout(layer_layout)
        return layer_group

    def get_labels_data(self) -> np.ndarray | None:
        """Get the input segmentation given the current selection in the
        layer dropdown.

        Returns:
            np.ndarray | None: The data of the labels layer with the name
                that is selected, or None if no layer is selected.
        """
        layer = self.layer_selection_box.value
        if layer is None:
            return None
        return layer.data

    def reshape_labels(self, segmentation: np.ndarray) -> None:
        """Expect napari segmentation to have shape t, [z], y, x.
        Motile toolbox needs a channel dimension between time and space.
        This also raises an error if the input seg is not the expected shape.

        Args:
            segmentation (np.ndarray): _description_

        Raises:
            ValueError if segmentation is not 3D or 4D.
        """
        ndim = segmentation.ndim
        if ndim > 4:
            raise ValueError(
                "Expected segmentation to be at most 4D, found %d", ndim
            )
        elif ndim < 3:
            raise ValueError(
                "Expected segmentation to be at least 3D, found %d", ndim
            )
        reshaped = np.expand_dims(segmentation, 1)
        return reshaped

    def _run_widget(self) -> QWidget:
        """Construct a widget where you set the run name and start solving.
        Initializes self.run_name and connects the generate tracks button
        to emit the new_run signal.

        Returns:
            QWidget: A row widget with a line edit for run name and a button
                to create a new run and start solving.
        """
        # Specify name text box
        run_widget = QWidget()
        run_layout = QHBoxLayout()
        run_layout.setContentsMargins(0, 0, 0, 0)
        run_layout.addWidget(QLabel("Run Name:"))
        self.run_name = QLineEdit("new_run")
        run_layout.addWidget(self.run_name)

        run_widget.setLayout(run_layout)
        return run_widget

    def get_run(self) -> MotileRun | None:
        """Construct a motile run from the current state of the run editor
        widget.

        Returns:
            MotileRun: A run with name, parameters, and input segmentation.
                Output segmentation and tracks not yet specified.
        """
        run_name = self.run_name.text()
        input_seg = self.get_labels_data()
        if input_seg is None:
            warn("No input labels layer selected", stacklevel=2)
            return None
        input_seg = self.reshape_labels(input_seg)
        params = self.solver_params_widget.solver_params.copy()
        return MotileRun(
            run_name=run_name,
            solver_params=params,
            input_segmentation=input_seg,
        )

    def emit_run(self) -> None:
        """Construct a run and start solving by emitting the start run
        signal for the main widget to connect to. If run is invalid, will
        not emit the signal.
        """
        run = self.get_run()
        if run is not None:
            self.start_run.emit(run)

    def new_run(self, run: MotileRun) -> None:
        """Configure the run editor to copy the name and params of the given
        run.
        """
        self.run_name.setText(run.run_name)
        self.solver_params_widget.new_params.emit(run.solver_params)



from functools import partial
from types import NoneType

from qtpy.QtCore import Signal
from qtpy.QtWidgets import (
    QCheckBox,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
    QWidget,
)

from motile_plugin.backend.solver_params import (
    CompoundSolverParam,
    SolverParams,
)

from .param_values import CompoundParamValue, EditableParamValue


class EditableParam(QWidget):
    def __init__(
        self,
        param_name: str,
        solver_params: SolverParams,
        negative: bool = False,
    ):
        """A widget for editing a parameter. Can be updated from
        the backend by calling update_from_params with a new SolverParams
        object. If changed in the UI, will emit a send_value signal which can
        be used to keep a SolverParams object in sync.

        Args:
            param_name (str): The name of the parameter to view in this UI row.
                Must correspond to one of the attributes of SolverParams.
            solver_params (SolverParams): The SolverParams object to use to
                initialize the view. Provides the title to display and the
                initial value.
            negative (bool, optional): Whether to allow negative values for
                this parameter. Defaults to False.
        """
        super().__init__()
        self.param_name = param_name
        field = solver_params.model_fields[param_name]
        self.dtype = field.annotation
        self.title = field.title
        self.negative = negative
        self.param_label = self._param_label_widget()
        self.param_label.setToolTip(field.description)
        self.param_value: CompoundParamValue | EditableParamValue
        if issubclass(CompoundSolverParam, self.dtype):
            self.param_value = CompoundParamValue(
                EditableParamValue(float, self.negative),
                EditableParamValue(float, self.negative),
            )
        else:
            self.param_value = EditableParamValue(float, self.negative)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.param_label)
        layout.addWidget(self.param_value)
        self.setLayout(layout)

        self.update_from_params(solver_params)

    def _param_label_widget(self) -> QLabel:
        return QLabel(self.title)

    def update_from_params(self, params: SolverParams):
        param_val = params.__getattribute__(self.param_name)
        if param_val is None:
            raise ValueError("Got None for required field {self.param_name}")
        else:
            self.param_value.update_value(param_val)


class OptionalEditableParam(EditableParam):
    def __init__(
        self,
        param_name: str,
        solver_params: SolverParams,
        negative: bool = False,
    ):
        """A widget for holding optional editable parameters. Adds a checkbox
        to the label, which toggles None-ness of the value.

        Args:
            param_name (str): _description_
            solver_params (SolverParams): _description_
            negative (bool, optional): _description_. Defaults to False.
        """
        super().__init__(param_name, solver_params, negative)
        self.param_label.toggled.connect(self.toggle_enable)

    def _param_label_widget(self) -> QCheckBox:
        return QCheckBox(self.title)

    def update_from_params(self, params: SolverParams):
        param_val = params.__getattribute__(self.param_name)
        if param_val is None:
            self.param_label.setChecked(False)
            self.param_value.setEnabled(False)
        else:
            self.param_label.setChecked(True)
            self.param_value.setEnabled(True)
            self.param_value.update_value(param_val)

    def toggle_enable(self, checked: bool):
        self.param_value.setEnabled(checked)
        value = self.param_value.get_value() if checked else None
        # force the parameter to say that the value has changed when we toggle
        self.param_value.valueChanged.emit(value)


class SolverParamsEditor(QWidget):
    """Widget for editing SolverParams.
    Spinboxes will be created for each parameter in SolverParams and linked such that
    editing the value in the spinbox will change the corresponding parameter.
    Checkboxes will also  be created for each optional parameter (group) and linked such
    that unchecking the box will update the parameter value to None, and checking will
    update the parameter to the current spinbox value.
    To update for a backend change to SolverParams, emit the new_params signal,
    which the spinboxes and checkboxes will connect to and use to update the
    UI and thus the stored solver params.
    """

    new_params = Signal(SolverParams)

    def __init__(self):
        super().__init__()
        self.solver_params = SolverParams()
        self.param_categories = {
            "hyperparams": ["max_edge_distance", "max_children"],
            "costs": [
                "appear_cost",
                "division_cost",
                "disappear_cost",
                "distance",
                "iou",
            ],
        }

        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.addWidget(
            self._params_group(
                "Hyperparameters", "hyperparams", negative=False
            )
        )
        main_layout.addWidget(
            self._params_group("Costs", "costs", negative=True)
        )
        self.setLayout(main_layout)

    def _params_group(
        self, title: str, param_category: str, negative: bool
    ) -> QWidget:
        widget = QGroupBox(title)
        layout = QVBoxLayout()
        # layout.setContentsMargins(0, 0, 0, 0)
        # layout.addWidget(QLabel(title))
        for param_name in self.param_categories[param_category]:
            field = self.solver_params.model_fields[param_name]
            param_cls = (
                OptionalEditableParam
                if issubclass(NoneType, field.annotation)
                else EditableParam
            )
            param_row = param_cls(
                param_name, self.solver_params, negative=negative
            )
            param_row.param_value.valueChanged.connect(
                partial(self.solver_params.__setattr__, param_name)
            )
            self.new_params.connect(param_row.update_from_params)
            layout.addWidget(param_row)
        widget.setLayout(layout)
        return widget



import networkx as nx
import numpy as np
import pytest
from motile_toolbox.candidate_graph.graph_attributes import EdgeAttr, NodeAttr
from skimage.draw import disk


@pytest.fixture
def segmentation_2d():
    frame_shape = (100, 100)
    total_shape = (2, *frame_shape)
    segmentation = np.zeros(total_shape, dtype="int32")
    # make frame with one cell in center with label 1
    rr, cc = disk(center=(50, 50), radius=20, shape=(100, 100))
    segmentation[0][rr, cc] = 1

    # make frame with two cells
    # first cell centered at (20, 80) with label 1
    # second cell centered at (60, 45) with label 2
    rr, cc = disk(center=(20, 80), radius=10, shape=frame_shape)
    segmentation[1][rr, cc] = 1
    rr, cc = disk(center=(60, 45), radius=15, shape=frame_shape)
    segmentation[1][rr, cc] = 2

    return np.expand_dims(segmentation, 1)


@pytest.fixture
def multi_hypothesis_segmentation_2d():
    """
    Creates a multi-hypothesis version of the `segmentation_2d` fixture defined above.

    """
    frame_shape = (100, 100)
    total_shape = (
        2,
        2,
        *frame_shape,
    )  # 2 time points, 2 hypotheses layers, H, W
    segmentation = np.zeros(total_shape, dtype="int32")
    # make frame with one cell in center with label 1 (hypo 1)
    rr0, cc0 = disk(center=(50, 50), radius=20, shape=frame_shape)
    # make frame with one cell at (45, 45) with label 1 (hypo 2)
    rr1, cc1 = disk(center=(45, 45), radius=15, shape=frame_shape)

    segmentation[0, 0][rr0, cc0] = 1
    segmentation[0, 1][rr1, cc1] = 1

    # make frame with two cells
    # first cell centered at (20, 80) with label 1
    rr0, cc0 = disk(center=(20, 80), radius=10, shape=frame_shape)
    rr1, cc1 = disk(center=(15, 75), radius=15, shape=frame_shape)

    segmentation[1, 0][rr0, cc0] = 1
    segmentation[1, 1][rr1, cc1] = 1

    # second cell centered at (60, 45) with label 2
    rr0, cc0 = disk(center=(60, 45), radius=15, shape=frame_shape)
    rr1, cc1 = disk(center=(55, 40), radius=20, shape=frame_shape)

    segmentation[1, 0][rr0, cc0] = 2
    segmentation[1, 1][rr1, cc1] = 2

    return segmentation


@pytest.fixture
def graph_2d():
    graph = nx.DiGraph()
    nodes = [
        (
            "0_1",
            {
                NodeAttr.POS.value: (50, 50),
                NodeAttr.TIME.value: 0,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
        (
            "1_1",
            {
                NodeAttr.POS.value: (20, 80),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
        (
            "1_2",
            {
                NodeAttr.POS.value: (60, 45),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_ID.value: 2,
            },
        ),
    ]
    edges = [
        (
            "0_1",
            "1_1",
            {EdgeAttr.IOU.value: 0.0},
        ),
        (
            "0_1",
            "1_2",
            {EdgeAttr.IOU.value: 0.395},
        ),
    ]
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    return graph


@pytest.fixture
def multi_hypothesis_graph_2d():
    graph = nx.DiGraph()
    nodes = [
        (
            "0_0_1",
            {
                NodeAttr.POS.value: (50, 50),
                NodeAttr.TIME.value: 0,
                NodeAttr.SEG_HYPO.value: 0,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
        (
            "0_1_1",
            {
                NodeAttr.POS.value: (45, 45),
                NodeAttr.TIME.value: 0,
                NodeAttr.SEG_HYPO.value: 1,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
        (
            "1_0_1",
            {
                NodeAttr.POS.value: (20, 80),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_HYPO.value: 0,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
        (
            "1_1_1",
            {
                NodeAttr.POS.value: (15, 75),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_HYPO.value: 1,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
        (
            "1_0_2",
            {
                NodeAttr.POS.value: (60, 45),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_HYPO.value: 0,
                NodeAttr.SEG_ID.value: 2,
            },
        ),
        (
            "1_1_2",
            {
                NodeAttr.POS.value: (55, 40),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_HYPO.value: 1,
                NodeAttr.SEG_ID.value: 2,
            },
        ),
    ]

    edges = [
        (
            "0_0_1",
            "1_0_1",
            {EdgeAttr.IOU.value: 0.0},
        ),
        (
            "0_0_1",
            "1_1_1",
            {EdgeAttr.IOU.value: 0.0},
        ),
        (
            "0_0_1",
            "1_0_2",
            {EdgeAttr.IOU.value: 0.3931},
        ),
        (
            "0_0_1",
            "1_1_2",
            {EdgeAttr.IOU.value: 0.4768},
        ),
        (
            "0_1_1",
            "1_0_1",
            {EdgeAttr.IOU.value: 0.0},
        ),
        (
            "0_1_1",
            "1_1_1",
            {EdgeAttr.IOU.value: 0.0},
        ),
        (
            "0_1_1",
            "1_0_2",
            {EdgeAttr.IOU.value: 0.2402},
        ),
        (
            "0_1_1",
            "1_1_2",
            {EdgeAttr.IOU.value: 0.3931},
        ),
    ]
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    return graph


def sphere(center, radius, shape):
    assert len(center) == len(shape)
    indices = np.moveaxis(np.indices(shape), 0, -1)  # last dim is the index
    distance = np.linalg.norm(
        np.subtract(indices, np.asarray(center)), axis=-1
    )
    mask = distance <= radius
    return mask


@pytest.fixture
def segmentation_3d():
    frame_shape = (100, 100, 100)
    total_shape = (2, *frame_shape)
    segmentation = np.zeros(total_shape, dtype="int32")
    # make frame with one cell in center with label 1
    mask = sphere(center=(50, 50, 50), radius=20, shape=frame_shape)
    segmentation[0][mask] = 1

    # make frame with two cells
    # first cell centered at (20, 50, 80) with label 1
    # second cell centered at (60, 50, 45) with label 2
    mask = sphere(center=(20, 50, 80), radius=10, shape=frame_shape)
    segmentation[1][mask] = 1
    mask = sphere(center=(60, 50, 45), radius=15, shape=frame_shape)
    segmentation[1][mask] = 2

    return np.expand_dims(segmentation, 1)


@pytest.fixture
def multi_hypothesis_segmentation_3d():
    """
    Creates a multi-hypothesis version of the `segmentation_3d` fixture defined above.

    """
    frame_shape = (100, 100, 100)
    total_shape = (2, 2, *frame_shape)  # 2 time points, 2 hypotheses
    segmentation = np.zeros(total_shape, dtype="int32")
    # make first frame with one cell in center with label 1
    mask = sphere(center=(50, 50, 50), radius=20, shape=frame_shape)
    segmentation[0, 0][mask] = 1
    mask = sphere(center=(45, 50, 55), radius=20, shape=frame_shape)
    segmentation[0, 1][mask] = 1

    # make second frame, first hypothesis with two cells
    # first cell centered at (20, 50, 80) with label 1
    # second cell centered at (60, 50, 45) with label 2
    mask = sphere(center=(20, 50, 80), radius=10, shape=frame_shape)
    segmentation[1, 0][mask] = 1
    mask = sphere(center=(60, 50, 45), radius=15, shape=frame_shape)
    segmentation[1, 0][mask] = 2

    # make second frame, second hypothesis with one cell
    # first cell centered at (15, 50, 70) with label 1
    # second cell centered at (55, 55, 45) with label 2
    mask = sphere(center=(15, 50, 70), radius=10, shape=frame_shape)
    segmentation[1, 1][mask] = 1

    return segmentation


@pytest.fixture
def graph_3d():
    graph = nx.DiGraph()
    nodes = [
        (
            "0_1",
            {
                NodeAttr.POS.value: (50, 50, 50),
                NodeAttr.TIME.value: 0,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
        (
            "1_1",
            {
                NodeAttr.POS.value: (20, 50, 80),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
        (
            "1_2",
            {
                NodeAttr.POS.value: (60, 50, 45),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_ID.value: 2,
            },
        ),
    ]
    edges = [
        ("0_1", "1_1"),
        ("0_1", "1_2"),
    ]
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    return graph


@pytest.fixture
def multi_hypothesis_graph_3d():
    graph = nx.DiGraph()
    nodes = [
        (
            "0_0_1",
            {
                NodeAttr.POS.value: (50, 50, 50),
                NodeAttr.TIME.value: 0,
                NodeAttr.SEG_HYPO.value: 0,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
        (
            "0_1_1",
            {
                NodeAttr.POS.value: (45, 50, 55),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_HYPO.value: 1,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
        (
            "1_0_1",
            {
                NodeAttr.POS.value: (20, 50, 80),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_HYPO.value: 0,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
        (
            "1_0_2",
            {
                NodeAttr.POS.value: (60, 50, 45),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_HYPO.value: 0,
                NodeAttr.SEG_ID.value: 2,
            },
        ),
        (
            "1_1_1",
            {
                NodeAttr.POS.value: (15, 50, 70),
                NodeAttr.TIME.value: 1,
                NodeAttr.SEG_HYPO.value: 1,
                NodeAttr.SEG_ID.value: 1,
            },
        ),
    ]
    edges = [
        ("0_0_1", "1_0_1"),
        ("0_0_1", "1_0_2"),
        ("0_1_1", "1_0_1"),
        ("0_1_1", "1_0_2"),
        ("0_0_1", "1_1_1"),
        ("0_1_1", "1_1_1"),
    ]
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    return graph



import logging
import shutil
import zipfile
from pathlib import Path
from urllib.request import urlretrieve

import tifffile
import zarr
from appdirs import AppDirs
from napari.types import LayerData

logger = logging.getLogger(__name__)


def Fluo_N2DL_HeLa() -> list[LayerData]:
    """Loads the Fluo-N2DL-HeLa 01 training raw data and silver truth from
    the appdir "user data dir". Will download it from the CTC and convert it to
    zarr if it is not present already.
    Returns:
        list[LayerData]: An image layer of 01 training raw data and a labels
            layer of 01 training silver truth labels
    """
    ds_name = "Fluo-N2DL-HeLa"
    appdir = AppDirs("motile-plugin")
    data_dir = Path(appdir.user_data_dir)
    data_dir.mkdir(parents=True, exist_ok=True)
    return read_ctc_dataset(ds_name, data_dir)


def read_ctc_dataset(ds_name: str, data_dir: Path) -> list[LayerData]:
    """Read a CTC dataset from a zarr (assumes pre-downloaded and converted)
    and returns a list of layer data for making napari layers

    Args:
        ds_name (str): Dataset name
        data_dir (Path): Path to the directory containing the zarr

    Returns:
        list[LayerData]: An image layer of 01 training raw data and a labels
            layer of 01 training silver truth labels
    """
    ds_zarr = data_dir / (ds_name + ".zarr")
    if not ds_zarr.exists():
        logger.info("Downloading %s", ds_name)
        download_ctc_dataset(ds_name, data_dir)

    store = zarr.NestedDirectoryStore(ds_zarr)
    raw_data = zarr.open(
        store=store,
        path="01",
    )[:]
    raw_layer_data = (raw_data, {"name": "01_raw"}, "image")
    seg_data = zarr.open(store=store, path="01_ST")[:]
    seg_layer_data = (seg_data, {"name": "01_ST"}, "labels")
    return [raw_layer_data, seg_layer_data]


def download_ctc_dataset(ds_name: str, data_dir: Path) -> None:
    """Download a dataset from the Cell Tracking Challenge
    and unzip it, then delete the zip. Then convert the tiffs to
    zarrs for the first training set images and silver truth.

    Args:
        ds_name (str): Dataset name, according to the CTC
        data_dir (Path): The directory in which to store the data.
    """
    ds_dir = data_dir / ds_name
    ds_zarr = data_dir / (ds_name + ".zarr")
    ctc_url = f"http://data.celltrackingchallenge.net/training-datasets/{ds_name}.zip"
    zip_filename = data_dir / f"{ds_name}.zip"
    if not zip_filename.is_file():
        urlretrieve(ctc_url, filename=zip_filename)
    with zipfile.ZipFile(zip_filename, "r") as zip_ref:
        zip_ref.extractall(data_dir)
    zip_filename.unlink()

    convert_to_zarr(ds_dir / "01", ds_zarr, "01")
    convert_to_zarr(ds_dir / "01_ST" / "SEG", ds_zarr, "01_ST")
    shutil.rmtree(ds_dir)


def convert_to_zarr(tiff_path: Path, zarr_path: Path, zarr_group: str):
    """Convert tiff file image data to zarr. Also deletes the tiffs!
    Args:
        tif_path (Path): Path to the directory containing the tiff files
        zarr_path (Path): path to the zarr file to write the output to
        zarr_group (Path): group within the zarr store to write the data to
    """
    # get data dimensions
    files = sorted(tiff_path.glob("*.tif"))
    logger.info("%s time points found.", len(files))
    example_image = tifffile.imread(files[0])
    data_shape = (len(files), *example_image.shape)
    data_dtype = example_image.dtype
    # prepare zarr
    zarr_path.mkdir(parents=True, exist_ok=True)
    store = zarr.NestedDirectoryStore(zarr_path)
    zarr_array = zarr.open(
        store=store,
        mode="w",
        path=zarr_group,
        shape=data_shape,
        dtype=data_dtype,
    )
    # load and save data in zarr
    for t, file in enumerate(files):
        image = tifffile.imread(file)
        zarr_array[t] = image
        file.unlink()
    tiff_path.rmdir()



import logging
import time

import numpy as np
from motile import Solver, TrackGraph
from motile.constraints import ExclusiveNodes, MaxChildren, MaxParents
from motile.costs import Appear, Disappear, EdgeDistance, EdgeSelection, Split
from motile_toolbox.candidate_graph import (
    EdgeAttr,
    NodeAttr,
    get_candidate_graph,
    graph_to_nx,
)

from .solver_params import SolverParams

logger = logging.getLogger(__name__)


def solve(
    solver_params: SolverParams,
    segmentation: np.ndarray,
    on_solver_update=None,
):
    cand_graph, conflict_sets = get_candidate_graph(
        segmentation,
        solver_params.max_edge_distance,
        iou=solver_params.iou is not None,
    )
    logger.debug("Cand graph has %d nodes", cand_graph.number_of_nodes())
    solver = construct_solver(cand_graph, solver_params, conflict_sets)
    start_time = time.time()
    solution = solver.solve(verbose=False, on_event=on_solver_update)
    logger.info("Solution took %.2f seconds", time.time() - start_time)

    solution_graph = solver.get_selected_subgraph(solution=solution)
    solution_nx_graph = graph_to_nx(solution_graph)

    return solution_nx_graph


def construct_solver(cand_graph, solver_params, exclusive_sets):
    solver = Solver(
        TrackGraph(cand_graph, frame_attribute=NodeAttr.TIME.value)
    )
    solver.add_constraints(MaxChildren(solver_params.max_children))
    solver.add_constraints(MaxParents(1))
    if exclusive_sets is None or len(exclusive_sets) > 0:
        solver.add_constraints(ExclusiveNodes(exclusive_sets))

    if solver_params.appear_cost is not None:
        solver.add_costs(Appear(solver_params.appear_cost))
    if solver_params.disappear_cost is not None:
        solver.add_costs(Disappear(solver_params.disappear_cost))
    if solver_params.division_cost is not None:
        solver.add_costs(Split(constant=solver_params.division_cost))

    if solver_params.distance is not None:
        solver.add_costs(
            EdgeDistance(
                position_attribute=NodeAttr.POS.value,
                weight=solver_params.distance.weight,
                constant=solver_params.distance.constant,
            ),
            name="distance",
        )
    if solver_params.iou is not None:
        solver.add_costs(
            EdgeSelection(
                weight=solver_params.iou.weight,
                attribute=EdgeAttr.IOU.value,
                constant=solver_params.iou.constant,
            ),
            name="iou",
        )
    return solver



import napari
import zarr
import argparse
import logging
from pathlib import Path
import sys
from darts_utils.data import (
    RawDataZarr,
    SegmentationZarr,
    add_data_args,
    add_segmentation_args,
)
import yaml
from motile_plugin import MotileWidget
from napari.utils.theme import _themes
from funlib.geometry.roi import Roi

_themes["dark"].font_size = "18pt"
logger = logging.getLogger(__name__)

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s %(name)s %(levelname)-8s %(message)s"
)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("config")
    args = parser.parse_args()

    with open(args.config) as f:
        try:
            config = yaml.safe_load(f)
        except yaml.YAMLError as exc:
            print(exc)

    storage_config = config["storage"]
    dataset_config = config["dataset"]    

    base_path = storage_config["data_base_path"]
    seg_path = storage_config["seg_base_path"]
    experiments_path = Path(storage_config["exp_base_path"])

    ds_name = dataset_config["dataset"]
    result_names = dataset_config["result_names"]
    roi = Roi(dataset_config["roi_offset"], dataset_config["roi_shape"])
    
    raw_zarr = RawDataZarr(base_path, ds_name, mode="r", store_type="flat")
    if not "fov" in dataset_config:
        fovs = raw_zarr.get_fovs()
        if not fovs:
            print(f"no fovs for dataset {ds_name}, exiting")
            sys.exit()
        fov = raw_zarr.get_fovs()[0]
    else:
        fov = dataset_config["fov"]

    if not "channels" in dataset_config:
        channels = raw_zarr.get_channels(fov)
    else:
        channels = dataset_config["channels"]
    seg_zarr = SegmentationZarr(
                seg_path, ds_name, None, mode="r"
            )
    
    print("Starting napari viewer")
    # Initialize Napari viewer
    viewer = napari.Viewer()
    for channel in channels:
        print(f"Adding images for channel {channel}")
        image_stack = raw_zarr.get_data(fov, channel)
        image_stack = image_stack[raw_zarr.roi_to_slices(roi)]
        viewer.add_image(image_stack[:, 0], name=f"{channel}_raw")

        fluorescent_prefixes = [
            "mCherry",
            "RFP",
            "YFP",
            "GFP",
            "CFP",
        ]
        if any(prefix in channel for prefix in fluorescent_prefixes):
            for result in result_names:
                seg_zarr.set_result_name(result)
                labeled_mask = seg_zarr.get_data(fov, channel)
                labeled_mask = labeled_mask[seg_zarr.roi_to_slices(roi)]
                viewer.add_labels(labeled_mask[:,0], name=f"{channel}_{result}")
            
    
    print("Done adding images")
    # Add your custom widget

    ds_exp_path = experiments_path  / ds_name
    ds_exp_path.mkdir(exist_ok=True)
    widget = MotileWidget(viewer, storage_path=ds_exp_path)
    viewer.window.add_dock_widget(widget)

    # Start the Napari GUI event loop
    napari.run()



from motile_plugin.backend.solve import solve
from motile_plugin.backend.solver_params import SolverParams


# capsys is a pytest fixture that captures stdout and stderr output streams
def test_solve_2d(segmentation_2d, graph_2d):
    params = SolverParams()
    params.appear_cost = None
    params.disappear_cost = None
    soln_graph = solve(params, segmentation_2d)
    assert set(soln_graph.nodes) == set(graph_2d.nodes)


def test_solve_3d(segmentation_3d, graph_3d):
    params = SolverParams()
    params.appear_cost = None
    params.disappear_cost = None
    soln_graph = solve(params, segmentation_3d)
    assert set(soln_graph.nodes) == set(graph_3d.nodes)



from functools import partial
from pathlib import Path
from warnings import warn

import pyqtgraph as pg
from fonticon_fa6 import FA6S
from motile_toolbox.candidate_graph import NodeAttr
from qtpy.QtCore import Signal
from qtpy.QtWidgets import (
    QFileDialog,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)
from superqt import QCollapsible
from superqt.fonticon import icon as qticon

from motile_plugin.backend.motile_run import MotileRun

from .params_viewer import SolverParamsViewer


class RunViewer(QGroupBox):
    """A widget for viewing in progress or completed runs, including
    the progress of the solver and the parameters. Can also save the whole
    run or export the tracks to CSV.
    Output tracks and segmentation are visualized separately in napari layers.
    """

    edit_run = Signal(object)

    def __init__(self):
        super().__init__(title="Run Viewer")
        # define attributes
        self.run: MotileRun = None
        self.params_widget = SolverParamsViewer()
        self.solver_label: QLabel
        self.gap_plot: pg.PlotWidget
        # Define persistent file dialogs for saving and exporting
        self.save_run_dialog = self._save_dialog()
        self.export_tracks_dialog = self._export_tracks_dialog()

        # Create layout and add subwidgets
        main_layout = QVBoxLayout()
        main_layout.addWidget(self._progress_widget())
        main_layout.addWidget(self.params_widget)
        main_layout.addWidget(self._save_and_export_widget())
        main_layout.addWidget(self._back_to_edit_widget())
        self.setLayout(main_layout)

    def update_run(self, run: MotileRun):
        self.run = run
        self.setTitle("Run Viewer: " + self._run_name_view(self.run))
        self.gap_plot.getPlotItem().clear()
        self.solver_event_update()
        self.params_widget.new_params.emit(run.solver_params)

    def _run_name_view(self, run: MotileRun) -> str:
        run_time = run.time.strftime("%m/%d/%y, %H:%M:%S")
        return f"{run.run_name} ({run_time})"

    def _save_and_export_widget(self):
        widget = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        # Save button
        icon = qticon(FA6S.floppy_disk, color="white")
        save_run_button = QPushButton(icon=icon, text="Save run")
        save_run_button.clicked.connect(self.save_run)
        layout.addWidget(save_run_button)

        # create button to export tracks
        export_tracks_btn = QPushButton("Export tracks to CSV")
        export_tracks_btn.clicked.connect(self.export_tracks)
        layout.addWidget(export_tracks_btn)

        widget.setLayout(layout)
        return widget

    def _back_to_edit_widget(self):
        widget = QWidget()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        # create back to editing button
        edit_run_button = QPushButton("Edit from previous state")
        edit_run_button.clicked.connect(partial(self.edit_run.emit, None))
        layout.addWidget(edit_run_button)

        # new run from this config button

        run_from_config_button = QPushButton("Edit from this state")
        run_from_config_button.clicked.connect(self._emit_run)
        layout.addWidget(run_from_config_button)
        widget.setLayout(layout)
        return widget

    def _emit_run(self):
        self.edit_run.emit(self.run)  # Note: this may cause memory leak
        # Can use weakref if that happens
        # https://github.com/Carreau/napari/commit/cd079e9dcb62de115833ea1b6bb1b7a0ab4b78d1

    def _progress_widget(self):
        progress_widget = QWidget()
        layout = QVBoxLayout()
        self.solver_label = QLabel("")
        self.gap_plot = self._plot_widget()
        collapsable_plot = QCollapsible("Graph of solver gap")
        collapsable_plot.layout().setContentsMargins(0, 0, 0, 0)
        collapsable_plot.addWidget(self.gap_plot)
        collapsable_plot.expand(animate=False)
        layout.addWidget(self.solver_label)
        layout.addWidget(collapsable_plot)
        layout.setContentsMargins(0, 0, 0, 0)
        progress_widget.setLayout(layout)
        return progress_widget

    def set_solver_label(self, status: str):
        message = "Solver status: " + status
        self.solver_label.setText(message)

    def _plot_widget(self) -> pg.PlotWidget:
        gap_plot = pg.PlotWidget()
        gap_plot.setBackground((37, 41, 49))
        styles = {
            "color": "white",
        }
        gap_plot.plotItem.setLogMode(x=False, y=True)
        gap_plot.plotItem.setLabel("left", "Gap", **styles)
        gap_plot.plotItem.setLabel("bottom", "Solver round", **styles)
        return gap_plot

    def plot_gaps(self, gaps: list[float]):
        if len(gaps) > 0:
            self.gap_plot.getPlotItem().plot(range(len(gaps)), gaps)
        else:
            self.gap_plot.getPlotItem().clear()

    def _save_dialog(self) -> QFileDialog:
        save_run_dialog = QFileDialog()
        save_run_dialog.setFileMode(QFileDialog.Directory)
        save_run_dialog.setOption(QFileDialog.ShowDirsOnly, True)
        return save_run_dialog

    def _export_tracks_dialog(self) -> QFileDialog:
        export_tracks_dialog = QFileDialog()
        export_tracks_dialog.setFileMode(QFileDialog.AnyFile)
        export_tracks_dialog.setAcceptMode(QFileDialog.AcceptSave)
        export_tracks_dialog.setDefaultSuffix("csv")
        return export_tracks_dialog

    def save_run(self):
        if self.save_run_dialog.exec_():
            directory = self.save_run_dialog.selectedFiles()[0]
            self.run.save(directory)
        else:
            warn("Saving aborted", stacklevel=2)

    def export_tracks(self):
        """Export the tracks from this run to a csv with the following columns:
        t,[z],y,x,id,parent_id
        Cells without a parent_id will have an empty string for the parent_id.
        Whether or not to include z is inferred from the length of an
        arbitrary node's position attribute.
        """
        default_name = MotileRun._make_directory(
            self.run.time, self.run.run_name
        )
        default_name = f"{default_name}_tracks.csv"
        base_path = Path(self.export_tracks_dialog.directory().path())
        self.export_tracks_dialog.selectFile(str(base_path / default_name))
        if self.export_tracks_dialog.exec_():
            outfile = self.export_tracks_dialog.selectedFiles()[0]
            header = ["t", "z", "y", "x", "id", "parent_id"]
            tracks = self.run.tracks
            _, sample_data = next(iter(tracks.nodes(data=True)))
            ndim = len(sample_data[NodeAttr.POS.value])
            if ndim == 2:
                header = [header[0]] + header[2:]  # remove z
            with open(outfile, "w") as f:
                f.write(",".join(header))
                for node_id, data in tracks.nodes(data=True):
                    parents = list(tracks.predecessors(node_id))
                    parent_id = "" if len(parents) == 0 else parents[0]
                    row = [
                        data[NodeAttr.TIME.value],
                        *data[NodeAttr.POS.value],
                        node_id,
                        parent_id,
                    ]
                    f.write("\n")
                    f.write(",".join(map(str, row)))
        else:
            warn("Exporting aborted", stacklevel=2)

    def solver_event_update(self):
        self.set_solver_label(self.run.status)
        self.plot_gaps(self.run.gaps)

    def reset_progress(self):
        self.set_solver_label("not running")
        self.gap_plot.getPlotItem().clear()



project = 'Motile Plugin'
copyright = '2024, Howard Hughes Medical Institute'
author = 'Caroline Malin-Mayor'

extensions = ['sphinx.ext.autodoc', 'myst_parser', 'autoapi.extension', 'sphinx_rtd_theme']
autoapi_dirs = ['../../src/motile_plugin']

exclude_patterns = []



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
# html_static_path = ['_static']



from qtpy.QtCore import Signal
from qtpy.QtWidgets import (
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
    QWidget,
)

from motile_plugin.backend.solver_params import (
    CompoundSolverParam,
    SolverParams,
)

from .param_values import CompoundParamValue, StaticParamValue


class ParamView(QWidget):
    def __init__(self, param_name: str, solver_params: SolverParams):
        """A widget for viewing a parameter (read only). Can be updated from
        the backend by calling update_from_params with a new SolverParams
        object.

        Args:
            param_name (str): The name of the parameter to view in this UI row.
                Must correspond to one of the attributes of SolverParams.
            solver_params (SolverParams): The SolverParams object to use to
                initialize the view. Provides the title to display and the
                initial value.
        """
        super().__init__()
        self.param_name = param_name
        field = solver_params.model_fields[param_name]
        self.dtype = field.annotation
        self.title = field.title
        self.param_label = QLabel(self.title)
        self.param_label.setToolTip(field.description)
        self.param_value: CompoundParamValue | StaticParamValue
        if issubclass(CompoundSolverParam, self.dtype):
            self.param_value = CompoundParamValue(
                StaticParamValue(), StaticParamValue()
            )
        else:
            self.param_value = StaticParamValue()

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.param_label)
        layout.addWidget(self.param_value)
        self.setLayout(layout)

        self.update_from_params(solver_params)

    def update_from_params(self, params: SolverParams):
        """Updates the current parameter value displayed in self.param_value,
        or hides the whole row if the value is None.

        Args:
            params (SolverParams): The solver parameters we use to update
                the UI. This object uses self.param_name to know which value
                to retrieve.
        """
        param_val = params.__getattribute__(self.param_name)
        if param_val is not None:
            self.param_value.update_value(param_val)
            self.show()
        else:
            self.hide()


class SolverParamsViewer(QWidget):
    """Widget for viewing SolverParams. To update for a backend change to
    SolverParams, emit the new_params signal, which each parameter label
    will connect to and use to update the UI.
    """

    new_params = Signal(SolverParams)

    def __init__(self):
        super().__init__()
        self.solver_params: SolverParams = SolverParams()
        self.param_categories = {
            "hyperparams": ["max_edge_distance", "max_children"],
            "costs": [
                "appear_cost",
                "division_cost",
                "disappear_cost",
                "distance",
                "iou",
            ],
        }
        main_layout = QVBoxLayout()
        main_layout.addWidget(
            self._params_group(
                title="Hyperparameters", param_category="hyperparams"
            )
        )
        main_layout.addWidget(
            self._params_group(title="Costs", param_category="costs")
        )
        main_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(main_layout)

    def _params_group(self, title: str, param_category: str) -> QGroupBox:
        """A helper function to create a group of parameters in the UI.
        Also connects each parameter to the self.new_params signal so the
        values can be updated when new parameters are viewed.

        Args:
            title (str): The title to put at the top of the group
            param_category (str): A key of self.param_categories used to
                get the list of parameters in this group.

        Returns:
            QGroupBox: A widget containing all the parameters in the group,
                with static titles and values that can be updated via emitting
                the self.new_params signal.
        """
        group = QGroupBox(title)
        layout = QVBoxLayout()
        for param_name in self.param_categories[param_category]:
            param_view = ParamView(param_name, self.solver_params)
            self.new_params.connect(param_view.update_from_params)
            layout.addWidget(param_view)
        group.setLayout(layout)
        return group



from rtree import index
from scipy.spatial import KDTree
import numpy as np

def build_tree(cylinders):
    """
    Build a Tree for 3D cylinders.

    :param cylinders: List of cylinders, where each cylinder is represented as ((x1, y1, z1), (x2, y2, z2), radius).
    :return: KDTree object.
    """
    p = index.Property()
    p.dimension = 3
    idx = index.Index(properties=p)
    bboxes = []
    
    for i, cylinder in enumerate(cylinders):
        p1, p2, radius = cylinder
        # Calculate the bounding box of the cylinder
        axis_vector = np.array(p2) - np.array(p1)
        axis_length = np.linalg.norm(axis_vector)

        if axis_length == 0:
            continue  # Skip degenerate cylinders

        axis_vector /= axis_length  # Normalize the axis vector
        orthogonal_vector = np.array([1, 0, 0]) if axis_vector[0] < 0.9 else np.array([0, 1, 0])
        ortho1 = np.cross(axis_vector, orthogonal_vector)
        ortho2 = np.cross(axis_vector, ortho1)

        corner_points = [p1, p2]
        for ortho in [ortho1, ortho2]:
            for sign in [-1, 1]:
                direction = ortho * radius * sign
                corner_points.extend([p1 + direction, p2 + direction])

        min_bounds = np.min(corner_points, axis=0)
        max_bounds = np.max(corner_points, axis=0)
        bbox = tuple(min_bounds) + tuple(max_bounds)
        bboxes += [bbox]

        # Insert the bounding box into the R-tree
        idx.insert(i, bbox)

    # Build and return the KD-Tree
    return idx, bboxes

# Example usage:
# cylinders = [((0,0,0), (1,1,1), 0.5), ((2,2,2), (3,3,3), 0.3)]
# tree = build_kdtree(cylinders)


# TODO dont hardcode graph
graph = viewer.layers[-1].data

cylinder_radius = 5
cylinders = []
for edge in graph.get_edges():
    (start_node, end_node) = edge[0]
    # cylinders += [(graph.get_coordinates()[start_node], graph.get_coordinates()[end_node], cylinder_radius)]
    coords = graph.get_coordinates(edge[0])
    
    cylinders += [(coords[0,:], coords[1,:],  cylinder_radius)]

tree, bboxes = build_tree(cylinders)




def ray_intersects_cylinder(ray_origin, ray_direction, cylinder, tolerance=1e-6):
    """
    Check if a ray intersects with a cylinder.

    :param ray_origin: Origin of the ray (x, y, z).
    :param ray_direction: Direction of the ray (dx, dy, dz).
    :param cylinder: Cylinder defined as ((x1, y1, z1), (x2, y2, z2), radius).
    :param tolerance: Numerical tolerance for the intersection test.
    :return: Boolean indicating if there is an intersection.
    """
    p1, p2, radius = cylinder
    d = np.array(p2) - np.array(p1)
    m = np.array(ray_origin) - np.array(p1)
    n = np.array(ray_direction)

    md = np.dot(m, d)
    nd = np.dot(n, d)
    dd = np.dot(d, d)

    # Coefficients for the quadratic equation
    a = dd * np.dot(n, n) - nd * nd
    b = dd * np.dot(n, m) - nd * md
    c = dd * np.dot(m, m) - md * md - radius * radius * dd

    # If a is approximately zero, the ray is parallel to the cylinder axis
    if abs(a) < tolerance:
        return False

    # Quadratic formula discriminant
    discr = b * b - a * c

    # If discriminant is negative, no real roots - no intersection
    if discr < 0:
        return False

    # Ray intersects cylinder
    return True

def distance_to_bbox(ray_origin, bbox):
    """
    Calculate the distance from a point to the closest point on a bounding box.

    :param ray_origin: Origin of the ray (x, y, z).
    :param bbox: Bounding box represented as a tuple (xmin, ymin, zmin, xmax, ymax, zmax).
    :return: Distance from the ray origin to the closest point on the bounding box.
    """
    xmin, ymin, zmin, xmax, ymax, zmax = bbox
    closest_point = np.maximum(np.minimum(ray_origin, [xmax, ymax, zmax]), [xmin, ymin, zmin])
    return np.linalg.norm(ray_origin - closest_point)


def query_ray_intersection(ray_origin, ray_direction, cylinders, rtree_index, bboxes):
    """
    Query the R-tree for a ray intersection with cylinders.

    :param ray_origin: Origin of the ray (x, y, z).
    :param ray_direction: Direction of the ray (dx, dy, dz).
    :param cylinders: List of cylinders.
    :param rtree_index: R-tree index object.
    :return: Index of the intersecting cylinder or None.
    """
    # Define a large bounding box along the ray for querying the R-tree
    ray_point_far = np.array(ray_origin) + np.array(ray_direction) * 10000  # Arbitrary large number
    bbox = tuple(np.minimum(ray_origin, ray_point_far)) + tuple(np.maximum(ray_origin, ray_point_far))

    # Query the R-tree for intersecting bounding boxes
    candidates = list(rtree_index.intersection(bbox))
    candidates.sort(key=lambda idx: distance_to_bbox(ray_origin, bboxes[idx]))
    
    for idx in candidates:
        if ray_intersects_cylinder(ray_origin, ray_direction, cylinders[idx]):
            return idx  # Return the first intersecting cylinder's index

    return None  # No intersection found

# Example usage:
ray_origin = (2, 367, 514)
ray_direction = (0, 1, 0)
result = query_ray_intersection(ray_origin, ray_direction, cylinders, tree, bboxes)

(result, cylinders[result])





from functools import partial
from warnings import warn

from napari._qt.qt_resources import QColoredSVGIcon
from qtpy.QtCore import Signal
from qtpy.QtWidgets import (
    QFileDialog,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QListWidget,
    QListWidgetItem,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from motile_plugin.backend.motile_run import MotileRun


class RunButton(QWidget):
    # https://doc.qt.io/qt-5/qlistwidget.html#setItemWidget
    # I think this means if we want static buttons we can just make the row here
    # but if we want to change the buttons we need to do something more complex
    # Lets start with static then!
    # Columns: Run name, Date/time, delete btn
    def __init__(self, run: MotileRun):
        super().__init__()
        self.run = run
        self.run_name = QLabel(self.run.run_name)
        self.run_name.setFixedHeight(20)
        self.datetime = QLabel(self.run.time.strftime("%m/%d/%y, %H:%M:%S"))
        self.datetime.setFixedHeight(20)
        icon = QColoredSVGIcon.from_resources("delete")
        self.delete = QPushButton(icon=icon.colored("white"))
        self.delete.setFixedSize(20, 20)
        layout = QHBoxLayout()
        layout.setSpacing(1)
        layout.addWidget(self.run_name)
        layout.addWidget(self.datetime)
        layout.addWidget(self.delete)
        self.setLayout(layout)

    def sizeHint(self):
        hint = super().sizeHint()
        hint.setHeight(30)
        return hint


class RunsList(QWidget):
    """Widget for holding in-memory Runs"""

    view_run = Signal(MotileRun)

    def __init__(self):
        super().__init__()
        self.runs_list: QListWidget
        self.file_dialog = QFileDialog()
        self.file_dialog.setFileMode(QFileDialog.Directory)
        self.file_dialog.setOption(QFileDialog.ShowDirsOnly, True)

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._tracking_runs_widget())
        self.setLayout(layout)

    def _selection_changed(self):
        selected = self.runs_list.selectedItems()
        if selected:
            self.view_run.emit(self.runs_list.itemWidget(selected[0]).run)

    def _tracking_runs_widget(self):
        save_load_group = QGroupBox("Tracking Runs")
        save_load_layout = QVBoxLayout()
        self.runs_list = QListWidget()
        self.runs_list.setSelectionMode(1)  # single selection
        self.runs_list.itemSelectionChanged.connect(self._selection_changed)
        save_load_layout.addWidget(self.runs_list)
        save_load_layout.addWidget(self._load_widget())
        save_load_group.setLayout(save_load_layout)
        return save_load_group

    def _load_widget(self):
        load_button = QPushButton("Load run")
        load_button.clicked.connect(self.load_run)
        return load_button

    def add_run(self, run: MotileRun, select=True):
        item = QListWidgetItem(self.runs_list)
        run_row = RunButton(run)
        self.runs_list.setItemWidget(item, run_row)
        item.setSizeHint(run_row.minimumSizeHint())
        self.runs_list.addItem(item)
        run_row.delete.clicked.connect(partial(self.remove_run, item))
        if select:
            self.runs_list.setCurrentRow(len(self.runs_list) - 1)

    def remove_run(self, item: QListWidgetItem):
        row = self.runs_list.indexFromItem(item).row()
        self.runs_list.takeItem(row)

    def load_run(self):
        if self.file_dialog.exec_():
            directory = self.file_dialog.selectedFiles()[0]
            try:
                run = MotileRun.load(directory)
                self.add_run(run, select=True)
            except (ValueError, FileNotFoundError) as e:
                warn(f"Could not load run from {directory}: {e}", stacklevel=2)



import json
from datetime import datetime
from pathlib import Path

import networkx as nx
import numpy as np
from pydantic import BaseModel

from .solver_params import SolverParams

STAMP_FORMAT = "%m%d%Y_%H%M%S"
PARAMS_FILENAME = "solver_params.json"
IN_SEG_FILEANME = "input_segmentation.npy"
OUT_SEG_FILEANME = "output_segmentation.npy"
TRACKS_FILENAME = "tracks.json"
GAPS_FILENAME = "gaps.txt"


class MotileRun(BaseModel):
    """An object representing a motile tracking run."""

    run_name: str
    solver_params: SolverParams
    input_segmentation: np.ndarray | None = None
    output_segmentation: np.ndarray | None = None
    tracks: nx.DiGraph | None = None
    time: datetime = datetime.now()
    gaps: list[float] = []
    status: str = "done"
    # pydantic does not check numpy arrays
    model_config = {"arbitrary_types_allowed": True}

    @staticmethod
    def _make_directory(time, run_name):
        stamp = time.strftime(STAMP_FORMAT)
        return f"{stamp}_{run_name}"

    @staticmethod
    def _unpack_directory(directory):
        stamp_len = len(datetime.now().strftime(STAMP_FORMAT))
        stamp = directory[0:stamp_len]
        run_name = directory[stamp_len + 1 :]
        try:
            time = datetime.strptime(stamp, STAMP_FORMAT)
        except ValueError as e:
            raise ValueError(
                f"Cannot unpack directory {directory} into timestamp and run name."
            ) from e
        return time, run_name

    def save(self, base_path: str | Path):
        base_path = Path(base_path)
        run_dir = base_path / self._make_directory(self.time, self.run_name)
        Path.mkdir(run_dir)
        self._save_params(run_dir)
        self._save_segmentation(
            run_dir, IN_SEG_FILEANME, self.input_segmentation
        )
        self._save_segmentation(
            run_dir, OUT_SEG_FILEANME, self.output_segmentation
        )
        self._save_tracks(run_dir)
        if self.gaps is not None:
            self._save_gaps(run_dir)

    @classmethod
    def load(cls, run_dir: Path | str):
        if isinstance(run_dir, str):
            run_dir = Path(run_dir)
        time, run_name = cls._unpack_directory(run_dir.stem)
        params = cls._load_params(run_dir)
        input_segmentation = cls._load_segmentation(run_dir, IN_SEG_FILEANME)
        output_segmentation = cls._load_segmentation(run_dir, OUT_SEG_FILEANME)
        tracks = cls._load_tracks(run_dir)
        gaps = cls._load_gaps(run_dir)
        return cls(
            run_name=run_name,
            solver_params=params,
            input_segmentation=input_segmentation,
            output_segmentation=output_segmentation,
            tracks=tracks,
            time=time,
            gaps=gaps,
        )

    def _save_params(self, run_dir):
        params_file = run_dir / PARAMS_FILENAME
        with open(params_file, "w") as f:
            json.dump(self.solver_params.__dict__, f)

    @staticmethod
    def _load_params(run_dir: Path) -> SolverParams:
        params_file = run_dir / PARAMS_FILENAME
        if not params_file.is_file():
            raise FileNotFoundError(f"Parameters not found at {params_file}")
        with open(params_file) as f:
            params_dict = json.load(f)
        return SolverParams(**params_dict)

    def _save_segmentation(self, run_dir, seg_file, segmentation):
        seg_file = run_dir / seg_file
        np.save(seg_file, segmentation)

    @staticmethod
    def _load_segmentation(
        run_dir: Path, seg_file: str, required: bool = True
    ) -> np.ndarray:
        seg_file = run_dir / seg_file
        if seg_file.is_file():
            return np.load(seg_file)
        elif required:
            raise FileNotFoundError(f"No segmentation at {seg_file}")
        else:
            return None

    def _save_tracks(self, run_dir: Path):
        tracks_file = run_dir / TRACKS_FILENAME
        with open(tracks_file, "w") as f:
            json.dump(nx.node_link_data(self.tracks), f)

    @staticmethod
    def _load_tracks(run_dir: Path, required: bool = True) -> nx.DiGraph:
        tracks_file: Path = run_dir / TRACKS_FILENAME
        if tracks_file.is_file():
            with open(tracks_file) as f:
                json_graph = json.load(f)
            return nx.node_link_graph(json_graph, directed=True)
        elif required:
            raise FileNotFoundError(f"No tracks at {tracks_file}")
        else:
            return None

    def _save_gaps(self, run_dir: Path):
        gaps_file = run_dir / GAPS_FILENAME
        with open(gaps_file, "w") as f:
            f.write(",".join(map(str, self.gaps)))

    @staticmethod
    def _load_gaps(run_dir, required: bool = True) -> list[float]:
        gaps_file = run_dir / GAPS_FILENAME
        if gaps_file.is_file():
            with open(gaps_file) as f:
                gaps = list(map(float, f.read().split(",")))
            return gaps
        elif required:
            raise FileNotFoundError(f"No gaps found at {gaps_file}")
        else:
            return None

    def delete(self, base_path: str | Path):
        base_path = Path(base_path)
        run_dir = base_path / self._make_directory(self.time, self.run_name)
        # Lets be safe and remove the expected files and then the directory
        (run_dir / PARAMS_FILENAME).unlink()
        (run_dir / IN_SEG_FILEANME).unlink()
        (run_dir / OUT_SEG_FILEANME).unlink()
        (run_dir / TRACKS_FILENAME).unlink()
        (run_dir / GAPS_FILENAME).unlink()
        run_dir.rmdir()



import numpy as np
import tifffile
import zarr
import os
from skimage.filters import threshold_otsu
from scipy.ndimage import label

# Step 1: Read the TIFF files
directory = "/Users/kharrington/git/cmalinmayor/motile-plugin/data/Fluo-N2DL-HeLa/01"
zarr_directory = "/Users/kharrington/git/cmalinmayor/motile-plugin/data/zarr_data.zarr"
file_names = sorted([f for f in os.listdir(directory) if f.endswith('.tif')])

# Step 2: Combine them into a 3D array
images = [tifffile.imread(os.path.join(directory, fname)) for fname in file_names]
stack = np.stack(images, axis=0)

# Step 3: Create a Zarr group and save the combined array
zarr_group = zarr.open_group(zarr_directory, mode='w')
zarr_group.create_dataset("stack", data=stack, chunks=(1, *stack.shape[1:]), dtype=stack.dtype)

# Step 4: Apply Otsu thresholding to create a segmentation mask
threshold_value = threshold_otsu(stack)
segmentation_mask = stack > threshold_value

# Step 5: Apply connected components to each slice of the segmentation mask
labeled_slices = [label(slice)[0] for slice in segmentation_mask]
labeled_stack = np.stack(labeled_slices, axis=0)

# Step 6: Save the segmentation mask and labeled stack in the Zarr group
zarr_group.create_dataset("segmentation_mask", data=segmentation_mask, chunks=(1, *segmentation_mask.shape[1:]), dtype=segmentation_mask.dtype)
zarr_group.create_dataset("labeled_stack", data=labeled_stack, chunks=(1, *labeled_stack.shape[1:]), dtype=labeled_stack.dtype)

print("Processing complete. Data saved in Zarr format.")



from importlib.metadata import PackageNotFoundError, version

from .widgets.motile_widget import MotileWidget

try:
    __version__ = version("motile-toolbox")
except PackageNotFoundError:
    # package is not installed
    __version__ = "uninstalled"

__all__ = ("MotileWidget",)



import logging

import napari
import zarr
from motile_plugin import MotileWidget
from napari.utils.theme import _themes

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s [%(filename)s:%(lineno)d] %(levelname)-8s %(message)s",
)
# logging.getLogger('napari').setLevel(logging.DEBUG)

_themes["dark"].font_size = "18pt"


# Load Zarr datasets
zarr_directory = "../data/Fluo-N2DL-HeLa.zarr"
zarr_group = zarr.open_group(zarr_directory, mode="r")

# Initialize Napari viewer
viewer = napari.Viewer()

# Add image and label layers to the viewer
viewer.add_image(zarr_group["01"][:], name="01 Raw")
viewer.add_labels(zarr_group["01_ST"][:], name="01 ST")

# Add your custom widget
widget = MotileWidget(viewer)
viewer.window.add_dock_widget(widget, name="Motile")

# Start the Napari GUI event loop
napari.run()



from pydantic import BaseModel, Field


class CompoundSolverParam(BaseModel):
    """A solver parameter with two values, a weight and a constant."""

    weight: float = Field(
        1,
        title="Weight",
        description=r"""Value multiplied by attribute to create the cost for selecting an edge.
The weight should be positive when the attribute is already a cost (higher is worse)
and negative when a higher value is better.
The magnitude of the weight helps balance this cost with other costs.""",
    )
    constant: float = Field(
        -20,
        title="Constant",
        description=r"""Value added to (attr * weight) to create the cost for selecting an edge.
A negative constant encourages selecting more edges.
(If all costs are positive, the optimal solution is selecting nothing.)""",
    )


class SolverParams(BaseModel):
    """The set of solver parameters supported in the motile plugin.
    Used to build the UI as well as store parameters for runs.
    """

    max_edge_distance: float = Field(
        50.0,
        title="Max Move Distance",
        description=r"""The maximum distance an object center can move between time frames.
Objects further than this cannot be matched, but making this value larger will increase solving time.""",
    )
    max_children: int = Field(
        2,
        title="Max Children",
        description="The maximum number of object in time t+1 that can be linked to an item in time t.\nIf no division, set to 1.",
    )
    appear_cost: float | None = Field(
        30,
        title="Appear",
        description=r"""Cost for starting a new track. A higher value means fewer selected tracks and fewer merges.""",
    )
    disappear_cost: float | None = Field(
        30,
        title="Disappear",
        description=r"""Cost for ending a track. A higher value means fewer selected tracks and fewer divisions.""",
    )
    division_cost: float | None = Field(
        20,
        title="Division",
        description=r"""Cost for a track dividing. A higher value means fewer divisions.
If this cost is higher than the appear cost, tracks will likely never divide.""",
    )
    distance: CompoundSolverParam | None = Field(
        CompoundSolverParam(weight=1, constant=-20),
        title="Distance",
        description=r"""Use the distance between objects as a feature for selecting tracks.""",
    )
    iou: CompoundSolverParam | None = Field(
        CompoundSolverParam(weight=5, constant=0),
        title="IOU",
        description=r"""Use the intersection over union between objects as a feature for selecting tracks.""",
    )



""" A module for widgets that hold parameter values. Each of these
widgets implements the following interface:

class ParamValue(QWidget):
    valueChanged = Signal(object)

    def update_value(self, value: Any) -> None:
        raise NotImplementedError()

    def get_value(self) -> Any:
        raise NotImplementedError

This "interface" is only enforced by duck typing, as is the pythonic way.
"""

from qtpy.QtCore import Signal
from qtpy.QtWidgets import QDoubleSpinBox, QGridLayout, QLabel, QWidget

from motile_plugin.backend.solver_params import CompoundSolverParam


class StaticParamValue(QLabel):
    """A widget for holding a parameter value (int or float) that cannot be
    changed from the UI. The valueChanged signal is just to fit my interface
    so that the compound widget can treat the static and editable children
    the same.
    """

    valueChanged = Signal(object)

    def update_value(self, value: int | float | None) -> None:
        if value is not None:
            text = str(value) if isinstance(value, int) else f"{value:.1f}"
            self.setText(text)

    def get_value(self) -> int | float:
        try:
            return int(self.text())
        except ValueError:
            return float(self.text())


class EditableParamValue(QDoubleSpinBox):
    valueChanged = Signal(object)

    def __init__(self, dtype: type, negative: bool = True) -> None:
        """A widget for holding an editable parameter value (int or float).
        The valueChanged signal is overriden from the QDoubleSpinbox
        to allow emitting NoneType.

        Args:
            dtype (type): The data type (int or float) of the parameter
            negative (bool, optional): Whether the value can be negative.
                Defaults to True.

        Raises:
            ValueError: If dtype is not (a superclass of) int or float.
        """
        super().__init__()
        if issubclass(int, dtype):
            self.setDecimals(0)
        elif issubclass(float, dtype):
            self.setDecimals(1)
        else:
            raise ValueError(f"Expected dtype int or float, got {dtype}")
        max_val = 10000
        min_val = -1 * max_val if negative else 0
        self.setRange(min_val, max_val)
        super().valueChanged.connect(self.valueChanged.emit)

    def update_value(self, value: int | float | None) -> None:
        if value is not None:
            self.setValue(value)

    def get_value(self) -> int | float:
        return self.value()


class CompoundParamValue(QWidget):
    valueChanged = Signal(object)

    def __init__(
        self,
        weight_widget: StaticParamValue | EditableParamValue,
        const_widget: StaticParamValue | EditableParamValue,
    ):
        """A widget to hold the value of a compound solver parameter (one that
        has weight and offset). This also implements the same interface as
        its component widgets for easy interchangability.

        Args:
            weight_widget (StaticParamValue | EditableParamValue): a widget
                holding the weight of the compound parameter
            offset_widget (StaticParamValue | EditableParamValue): a widget
                holding the offset of the compound parameter
        """
        super().__init__()
        self.weight = weight_widget
        self.constant = const_widget
        self.weight.valueChanged.connect(
            lambda: self.valueChanged.emit(self.get_value())
        )
        self.constant.valueChanged.connect(
            lambda: self.valueChanged.emit(self.get_value())
        )

        layout = QGridLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        weight_label = QLabel("weight:")
        weight_label.setToolTip(
            CompoundSolverParam.model_fields["weight"].description
        )
        layout.addWidget(weight_label, 0, 0)
        layout.addWidget(self.weight, 0, 1)
        constant_label = QLabel("constant:")
        layout.addWidget(constant_label, 1, 0)
        constant_label.setToolTip(
            CompoundSolverParam.model_fields["constant"].description
        )
        layout.addWidget(self.constant, 1, 1)
        self.setLayout(layout)

    def update_value(self, value: CompoundSolverParam) -> None:
        self.weight.update_value(value.weight)
        self.constant.update_value(value.constant)

    def get_value(self) -> CompoundSolverParam:
        return CompoundSolverParam(
            weight=self.weight.get_value(), constant=self.constant.get_value()
        )
